"Shuts out std streams from console"
import os
import sys

class NoStdStreams(object):
    def __init__(self,stdout = None, stderr = None):
        self.devnull = open(os.devnull,'w')
        self._stdout = stdout or self.devnull or sys.stdout
        self._stderr = stderr or self.devnull or sys.stderr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush(); self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush(); self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        self.devnull.close()



from .no_std_streams import NoStdStreams
from .functions import PrintSelf, SaveObj, LoadObj
from .logger import Logger
from .matplotlibstyle import *
__all__ = [
    'NoStdStreams',
    'PrintSelf',
    'SaveObj',
    'LoadObj',
    'Logger',
]



"Cell related functions"
from semiconductor.recombination import SRH
from ..utils import NoStdStreams
import scipy.constants as const
import numpy as np
import warnings


class Cell():
    #****   Constant declaration    ****#
    kb = const.physical_constants['Boltzmann constant in eV/K'][0]
    Tmin = 75
    Tmax = 400
    Ndopmin = 1e10
    Ndopmax = 1e20
    DefaultNdop = 1E15
    DefaultType = "p"
    DefaultT=300

    #****   Method declaration      ****#
    def __init__(self,T=None,Ndop=None,type=None,):
        '''
        ---Doc---
            Description:
                Validates Temperature range (Tmin-Tmax), and cell-type, defaulting to p-type
                Calculates Bang Gap Narrowing intrinsic carrier ni
                Calculates thermal velocity vn, vp
                Calculate n0,p0

                Calculation are made using MK8J semiconductor package :
                https://github.com/MK8J/semiconductor

                Models used:
                    - Band Gap Narrowing : Yan, 2014
                    - Intrinsici carrier : Couderc 2014 doi- 10.1063/1.4867776
                    - Thermal velocity : Green 1990 doi- 10.1063/1.345414
                    - Bandgap : Passler 2002 doi- 10.1103/PhysRevB.66.085201

            Inputs:
                T       Float       Temperature of the cell in Kelvin
                Ndop    Float       Doping level of the sample
                type    String      Cell type. Have "n" or "p"

            Outputs:
                object  represents defined cell

            Exemple:
                >>  Cell(T=300, Ndop = 1E15, type="p")
        '''
        if Ndop is None: Ndop = Cell.DefaultNdop
        if type is None: type = Cell.DefaultType
        if T is None: T = Cell.DefaultT

        #   Check dopping range
        if Ndop<Cell.Ndopmin:
            self.Ndop = Cell.Ndopmin
            warnings.warn("In Cell.__init__ : Ndop value out of bound, got %s, and expected within [%s,%s]. Et will be clipped to %s." %(Ndop,Cell.Ndopmin,Cell.Ndopmax,Cell.Ndopmin))
        elif Ndop>Cell.Ndopmax:
            self.Ndop = Cell.Ndopmax
            warnings.warn("In Cell.__init__ : Ndop value out of bound, got %s, and expected within [%s,%s]. Et will be clipped to %s." %(Ndop,Cell.Ndopmin,Cell.Ndopmax,Cell.Ndopmax))
        else:
            self.Ndop = Ndop

        #   Check temperature range
        if T<Cell.Tmin:
            self.T = Cell.Tmin
            warnings.warn("In Cell.__init__ : T value out of bound, got %s, and expected within [%s,%s]. Et will be clipped to %s." %(T,Cell.Tmin,Cell.Tmax,Cell.Tmin))
        elif T>Cell.Tmax:
            self.T = Cell.Tmax
            warnings.warn("In Cell.__init__ : T value out of bound, got %s, and expected within [%s,%s]. Et will be clipped to %s." %(T,Cell.Tmin,Cell.Tmax,Cell.Tmax))
        else:
            self.T = T

        #   Check type validity
        if type not in ["p","n"]:
            warnings.warn("In Cell.__init__ : Incorrect type value, expected n or p. Value will be defaulted to %s"%(Cell.DefaultType))
            self.type = Cell.DefaultType
        else:
            self.type = type

        #   Calculate Ni, Vn using semiconductor package
        with NoStdStreams():
            if type == "p": Tmodel=SRH(material="Si",temp = self.T, Nt = 1, nxc = 1e15, Na = self.Ndop, Nd= 0, BGN_author = "Yan_2014fer")
            if type == "n": Tmodel=SRH(material="Si",temp = self.T, Nt = 1, nxc = 1e15, Na = 0, Nd= self.Ndop, BGN_author = "Yan_2013fer")
        self.ni = Tmodel.nieff[0]
        self.Vn = Tmodel.vel_th_e[0]
        self.Vp = Tmodel.vel_th_h
        #Calculate n0 and p0
        if type == "p":
            self.p0 = (0.5 * (np.abs(Ndop - 0) + np.sqrt((0 - Ndop)**2 + 4 * self.ni**2)))
            self.n0 = (self.ni**2)/self.p0
        if type == "n":
            self.n0 = (0.5 * (np.abs(Ndop - 0) + np.sqrt((0 - Ndop)**2 + 4 * self.ni**2)))
            self.p0 = (self.ni**2)/self.n0
    def changeT(self,T):
        '''
        ---Doc---
            Description:
                Create a new cell instance, based on the same parameters as .self except for the new inputted temperature.

            Inputs:
                T       Float       Temperature of the cell in Kelvin

            Outputs:
                object  represents defined cell

            Exemple:
                >>  cell2 = Cell(T = 300, Ndop = 1E15, type="p").changeT(350)
        '''
        return Cell(T,self.Ndop,self.type)
    def changeNdop(self,Ndop):
        '''
        ---Doc---
            Description:
                Create a new cell instance, based on the same parameters as .self except for the new inputted doping.

            Inputs:
                Ndop       Float       Doping level of the sample

            Outputs:
                object  represents defined cell

            Exemple:
                >>  cell2 = Cell(T = 300, Ndop = 1e15, type="p").changeNdop(1e16)
        '''
        return Cell(self.T,Ndop,self.type)



import numpy as np
import warnings

class LTS():
    #****   Constant declaration    ****#
    N0_eeh=3.3e17   #cm-3
    N0_ehh =7.0e17 #cm-3
    B_temp = [77,90,112,170,195,249,300]    #K
    B_values = [8.01e-14,4.57e-14,2.14e-14,8.84e-15,7.35e-15,5.48e-15,4.73e-15] #cm3/s
    #****   Method declaration      ****#
    def __init__(self,cell,defect,dnrange,noise="", noiseparam=0):
        '''
        ---Doc---
            Description:
                Defines the boundary of the lifetime simulation, using the cell and defect parameters to calculate SRH lifetime.
                It is spossible to add noise to the lifetime, according to different noise models.

            Inputs:
                cell        object      Cell object previously created
                defect      object      Defect object previously created
                dnrange     array       Excess carrier concentration values at which to simulate data
                noise       String      Define which noise model to used
                noiseparam  float       hyper-parameter for noise level

            Outputs:
                Object

            Exemple:
                >> LTS(cell(300,1E15,"p"),defect(0.33,1E-15,4E-14,1E12),[1E-13,1E-15,1E-17], noise="log-gaussian")
        '''
        self.cell = cell
        self.defect = defect
        self.dnrange = dnrange
        self.n1 = self.cell.ni*np.exp(self.defect.Et/(self.cell.kb*self.cell.T))
        self.p1 = self.cell.ni*np.exp(-self.defect.Et/(self.cell.kb*self.cell.T))
        self.taun0 = 1/(self.defect.Sn * self.cell.Vn*self.defect.Nt)
        self.taup0 = 1/(self.defect.Sp * self.cell.Vp*self.defect.Nt)
        self.tauSRH = [(self.taun0*(self.cell.p0+self.p1+dn)+self.taup0*(self.cell.n0+self.n1+dn))/(self.cell.n0+self.cell.p0+dn) for dn in self.dnrange]

        #   Add noise
        self.noisemodel=noise
        self.noiseparam=noiseparam
        if self.noisemodel=="logNorm":
            noise = np.random.normal(0,noiseparam,len(dnrange))
            self.tauSRH_noise = [t*(1+e*np.log(np.max(self.dnrange)/dn)) for t,e,dn in zip(self.tauSRH,noise,self.dnrange)]
            self.noisemodel="logNorm"
            self.noiseparam=noiseparam
        else:
            self.tauSRH_noise=self.tauSRH
    def checkAuger(self):
        '''
        ---Doc---
            Description:
                Return a boolean to check if generated lifetime is smaller than Auger lifetime and a clip lifetime to Auger if needed
                Auger Constants are defined from Richter et al [2012] -DOI 10.1103/PhysRevB.86.165202
                Radiative contants from Altermat et al [2005] - DOI 10.1109/NUSOD.2005.1518128 and Trupke et al [2003] - DOI 10.1063/1.1610231

            Inputs:
                cell        object      Cell object previously created
                defect      object      Defect object previously created
                dnrange     array       Excess carrier concentration values at which to simulate data
                noise       String      Define which noise model to used
                noiseparam  float       hyper-parameter for noise level

            Outputs:
                breakAuger  boolean     True if lifetime is higher than Auger limit at any carrier concentration
                tauAuger    array       Auger lifetime array, the length of self.tauSRH
        '''
        g_eeh = 1+13*(1-np.tanh(np.power(self.cell.n0/LTS.N0_eeh,0.66)))
        g_ehh = 1+13*(1-np.tanh(np.power(self.cell.n0/LTS.N0_ehh,0.63)))
        tauAuger = np.zeros(len(self.tauSRH))
        bmin = 0.20+(0.00-0.20)/(1+np.power(self.cell.T/320,2.5))
        b1 = 1.5e18+(1e7-1.5e18)/(1+np.power(self.cell.T/550,3))
        b3 = 4e18+(1e9-4e18)/(1+np.power(self.cell.T/365,3.54))
        for k in range(len(LTS.B_temp)):
            if self.cell.T < LTS.B_temp[k]:
                if k==0: B_low = B_V[0]
                B_low=(LTS.B_values[k]-LTS.B_values[k-1])/(LTS.B_temp[k]-LTS.B_temp[k-1])*(self.cell.T-LTS.B_temp[k-1])+LTS.B_values[k-1]
            else:
                if k==len(LTS.B_temp)-1: B_low = LTS.B_values[-1]
        breakAuger=False
        for i in range(len(self.dnrange)):
            dn = self.dnrange[i]
            fact1 = (self.cell.n0+dn)*(self.cell.p0+dn)-self.cell.ni**2
            fact2 = 2.5e-31*g_eeh*self.cell.n0+8.5e-32*g_ehh*self.cell.p0+3e-29*np.power(dn,0.92)
            factB = B_low*(bmin+(1-bmin)/(1+np.power((self.cell.n0+dn+self.cell.p0+dn)/b1,0.54)+np.power((self.cell.n0+dn+self.cell.p0+dn)/b3,1.25)))
            tauAuger[i]=dn/(fact1*(fact2+factB))
            if tauAuger[i]<self.tauSRH[i]: breakAuger=True

        return(breakAuger,tauAuger)



"Write to a log file everything passed to print()"
import sys
import warnings
import numpy as np
class Logger():
    """Write to a log file everything passed to print().

    Parameters
    ----------
    logfile : str
        path to the log file.

    Examples
    -------
    Define the logger
    >>> logger = Logger(logfile)

    Start logging sys.stdout
    >>> logger.open()

    End logging sys.stdout
    >>> logger.close()

    Attributes
    ----------
    terminal :
        local storage of original sys.stdout.
    log :
        file handler for logfile.

    """
    #****   Constant declaration    ****#
    TitleLength=40

    #****   Core methods    ****#
    def __init__(self, logfile: str):
        self.terminal=None
        self.log = None
        self.logfile=logfile
    def __getattr__(self, attr):
            return getattr(self.terminal, attr)
    def write(self, message):
        """Overwrite write method to enable writing in both sys.stdout and log file."""
        self.terminal.write(message)
        self.terminal.flush()
        self.log.write(message)
    def close(self):
        """Close log file and restore sys.stdout"""
        self.log.close()
        sys.stdout = self.terminal
    def open(self):
        """Open log file and save sys.stdout"""
        self.terminal = sys.stdout
        self.log = open(self.logfile, "a+")
        sys.stdout = self

    #****   Additionnal methods related to printing    ****#
    def printTitle(title, titleLen = None, newLine=True):
        """Print out title header"""
        if titleLen == None: titleLen = Logger.TitleLength
        if len(title)>titleLen: warnings.warn("In makeTitle, title length bigger than %s"%(titleLen))
        print("="*np.max([0,np.int((titleLen-len(title))/2)+(titleLen-len(title))%2])+" "+title+" "+"="*np.max([0,np.int((titleLen-len(title))/2)]))
        if newLine: print('\n')
    def printDic(dic,skipKeys=None):
        """Print dictionary nicely, skipping the 'skipKeys' keys"""
        if skipKeys==None: skipKeys=['']
        for k in dic:
            if k in skipKeys: continue
            print("\t",k,"-"*(1+len(max(dic,key=len))-len(k)),">",dic[k])
        print('\n')



"Defines matplotlib stylesheet"
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

# %%--  Matplotlib style sheet
mpl.style.use('seaborn-paper')
mpl.rcParams['font.family'] = 'serif'
mpl.rcParams['font.serif'] ='STIXGeneral'
mpl.rcParams['font.size'] = 14
mpl.rcParams['mathtext.default'] = 'rm'
mpl.rcParams['mathtext.fallback_to_cm'] = False
mpl.rcParams['mathtext.fontset'] = 'stix'
mpl.rcParams['axes.labelsize'] = 16
mpl.rcParams['axes.labelweight'] = 'normal'
mpl.rcParams['axes.grid.which']='both'
mpl.rcParams['axes.xmargin']=0.05
mpl.rcParams['axes.ymargin']=0.05
mpl.rcParams['grid.linewidth']= 0.5
mpl.rcParams['xtick.labelsize'] = 14
mpl.rcParams['xtick.bottom'] = True
mpl.rcParams['xtick.top'] = True
mpl.rcParams['xtick.direction'] = 'in'
mpl.rcParams['ytick.left'] = True
mpl.rcParams['ytick.right'] = True
mpl.rcParams['ytick.direction'] = 'in'
mpl.rcParams['ytick.labelsize'] = 14
mpl.rcParams['legend.fontsize'] = 14
mpl.rcParams['figure.titlesize'] = 18
mpl.rcParams['figure.figsize'] = (8.09,5)
mpl.rcParams['figure.autolayout'] = False
mpl.rcParams['figure.dpi'] = 75
mpl.rcParams['image.cmap'] = "viridis"
mpl.rcParams['savefig.dpi'] = 150
mpl.rcParams['errorbar.capsize'] = 3
mpl.rcParams['axes.prop_cycle'] = plt.cycler(color = plt.cm.viridis([0.8,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9]))
# %%-



'Defect parameters and generation functions'
import numpy as np
import warnings

class Defect():
    #****   Constant declaration    ****#
    Etbound = 0.6       #   Absolute bound for Defect level
    Sminbound = 1E-20   #   Absolute minimum bound for Capture Cross Section
    Smaxbound = 1E-10   #   Absolute maximum bound for Capture Cross Section
    DefaultNt = 1E12    #   Default value for Defect density if none is provided

    #****   Method declaration      ****#
    def __init__(self,Et,Sn,Sp,Nt=None, name=""):
        '''
        ---Doc---
            Description:
                Take Defect parameters as inputs.
                Note that Et is defined as Et-Ei with Ei = Eg/2 and must therefore be between -Defect.Etbound and Defect.Etbound
                if outside the range, will clip to the side and throw a warning.
                Electron and hole capture-cross section are also bound between Sminbount and Smaxbound and will be clipped if exceeded.
                Calculates An and Ap: An = 1/(Sn*Nt) ;; Ap = 1/(Sp*Nt)
                Calculate k = Sn/Sp = Ap/An

            Inputs:
                Et  Float       Defect energy level, relative to the intrinsic mid-gap
                Sn  Float       Capture-cross section for electrons
                Sp  Float       Capture-cross section for holes
                Nt  Float       Defect density

            Outputs:
                object  represents defined Defect

            Exemple:
                >>  myDefect=Defect(0.33,1E-14,3E-15, 1E12)
        '''
        if Nt is None : Nt = Defect.DefaultNt
        if Et<-Defect.Etbound:
            self.Et = -Defect.Etbound
            warnings.warn("In Defect.__init__ : Et value out of bound, got %s, and expected within [%s,%s]. Et will be clipped to %s." %(Et,-Defect.Etbound,Defect.Etbound,-Defect.Etbound))
        elif Et>Defect.Etbound:
            self.Et = Defect.Etbound
            warnings.warn("In Defect.__init__ : Et value out of bound, got %s, and expected within [%s,%s]. Et will be clipped to %s." %(Et,-Defect.Etbound,Defect.Etbound,Defect.Etbound))
        else:
            self.Et=Et

        if Sn<Defect.Sminbound:
            self.Sn = Defect.Sminbound
            warnings.warn("In Defect.__init__ : Sn value out of bound, got %s, and expected within [%s,%s]. Sn will be clipped to %s." %(Sn,Defect.Sminbound,Defect.Smaxbound,Defect.Sminbound))
        elif Sn>Defect.Smaxbound:
            self.Sn = Defect.Smaxbound
            warnings.warn("In Defect.__init__ : Sn value out of bound, got %s, and expected within [%s,%s]. Sn will be clipped to %s." %(Sn,Defect.Sminbound,Defect.Smaxbound,Defect.Sminbound))
        else:
            self.Sn=Sn

        if Sp<Defect.Sminbound:
            self.Sp = Defect.Sminbound
            warnings.warn("In Defect.__init__ : Sp value out of bound, got %s, and expected within [%s,%s]. Sp will be clipped to %s." %(Sp,Defect.Sminbound,Defect.Smaxbound,Defect.Sminbound))
        elif Sp>Defect.Smaxbound:
            self.Sp = Defect.Smaxbound
            warnings.warn("In Defect.__init__ : Sp value out of bound, got %s, and expected within [%s,%s]. Sp will be clipped to %s." %(Sp,Defect.Sminbound,Defect.Smaxbound,Defect.Sminbound))
        else:
            self.Sp=Sp

        self.Nt=Nt
        self.An=1/(Sn*Nt)
        self.Ap=1/(Sp*Nt)
        self.k=Sn/Sp
        self.name=name
    def randomDB(N,Et_min = None, Et_max = None, S_min = None, S_max = None, Nt = None):
        '''
        ---Doc---
            Description:
                Generates N Defects objects bound by the imputted limits and store them in an array.
            Inputs:
                N           Integer     Number of Defects to generate in the database
                Et_min      Float       Lower boundary for Defect Energy level
                Et_max      Float       Higher boundary for Defect Energy level
                S_min       Float       Lower boundary for Capture Cross-Section
                S_max       Float       Higher boundary for Capture Cross-Section
                Nt          Float       Density of Defects
            Outputs:
                Res         array       Database of Defects
            Exemple:
                >>  Defect.random_db(100)
        '''
        #   If values are not define, then default thermal
        if Et_min is None: Et_min = - Defect.Etbound
        if Et_max is None: Et_max =  Defect.Etbound
        if S_min is None: S_min =  Defect.Sminbound
        if S_max is None: S_max =  Defect.Smaxbound
        if Nt is None: Nt = Defect.DefaultNt
        Res=[]
        for i in range(N):
            Et = np.random.rand(1)[0]*(Et_max-Et_min)+Et_min
            Sn = np.exp(np.random.rand(1)[0]*(np.log(S_max)-np.log(S_min))+np.log(S_min))
            Sp= np.exp(np.random.rand(1)[0]*(np.log(S_max)-np.log(S_min))+np.log(S_min))
            d = Defect(Et, Sn, Sp, Nt=Nt,name="D-"+str(i).zfill(1+int(np.trunc(np.log10(N))))) #    Will display leading '0' for naming ordering purposes
            Res.append(d)
        return Res



"Machine learning functions associated to SRH parameter extraction"
from ..utils.matplotlibstyle import *
from ..utils import SaveObj, LoadObj
from ..utils import Logger
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import datetime
import time
from sklearn.model_selection import *
from sklearn.preprocessing import *
from sklearn.metrics import *
from sklearn.ensemble import *
from sklearn.tree import *
from sklearn.neural_network import *
from sklearn.linear_model import *
from sklearn.svm import *
from sklearn.neighbors import *
from sklearn.model_selection import *

class ML():
    #****   Constant declaration    ****#
    DefaultParameters = {
        'name':'',
        'save': False,
        'logML': False,
        'mlID':None,
    }

    #****   Core methods      ****#
    def __init__(self, Dataset, SaveDir, Parameters=None):
        '''
        ---Doc---
            Description:
                Initialize ml object with passed or default parameters
            Inputs:
                Dataset     object      Dataset used for training
                SaveDir     string      Folder path to save the data if parameters['save'] is true
                Parameters  dicitionary Force or overwrite default parameters
            Outputs:
                None
        '''
        #   Use default if not defined
        self.parameters=ML.DefaultParameters
        if Parameters is not None: self.updateParameters(Parameters)

        #   Create directory for computation on this dataset
        self.pathDic = {
            'savedir':      SaveDir,
            'figures':      SaveDir+"\\figures\\",
            'objects':      SaveDir+"\\objects\\",
            'traces':       SaveDir+"\\traces\\",
            'outputs':      SaveDir+"\\outputs\\",
        }
        for key, value in self.pathDic.items():
                if key in ['figures','objects','traces','outputs']:
                    if not os.path.exists(value):   os.makedirs(value)
        #   define hyper parameters for ML training
        self.dataset = Dataset.copy(deep=True)
        self.logTrain={}
        self.logger = None
        self.pathDic['logfile']=self.pathDic['traces']+datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")+"_"+self.parameters['name']+".txt"
        if self.parameters['logML']: self.logger = Logger(self.pathDic['logfile'])

        #   Print header for logfile
        if self.parameters['logML']: self.logger.open()
        Logger.printTitle(self.parameters['name'],titleLen=80, newLine=False)
        Logger.printTitle('HYPER-PARAMETERS',titleLen=60, newLine=False)
        Logger.printTitle('PARAMETERS',titleLen=20)
        Logger.printDic(self.parameters)
        Logger.printTitle('PATH',titleLen=20)
        Logger.printDic(self.pathDic)
        if self.parameters['logML']: self.logger.close()
    def saveML(self, name=None):
        '''
        ---Doc---
            Description:
                Save ml object with pickle
            Inputs:
                name     string      Overwrite filename
            Outputs:
                None
        '''
        if name == None: name = self.parameters['name']
        if self.logger != None: self.logger = self.logger.logfile
        SaveObj(self,self.pathDic['objects'],'mlObj_'+name+"_"+datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S"))
    def updateParameters(self,Parameters):
        '''
        ---Doc---
            Description:
                update parameters dicitionary
            Inputs:
                Parameters  dicitionary Parameters to overwrite
            Outputs:
                None
        '''
        for key,value in Parameters.items():
            self.parameters[key]=value

    #****   ML methods      ****#
    def trainRegressor(self, targetCol, trainParameters=None):
        '''
        ---Doc---
            Description:
                training pipeline for regressors. Data and results are stored on the ML object
            Inputs:
                targetCol       string      column name of the dataframe to predict
                trainParameters dicitionary training parameters to overwrite
            Outputs:
                None
        '''
        #   define and update training parameters:
        trainParam={
            'validation_fraction': 0.01,    # validation dataset percentage
            'normalize': True,     # Wether or not to normalize the input data (True for NN)
            'base_model': RandomForestRegressor(n_estimators=100, n_jobs=-1, verbose=0),
            'random_seed': np.random.randint(1000),
            'bandgap': 'all', #or 'upper' == Et>0 or 'lower' ==Et<0
            }
        if trainParameters!=None:
            for key in trainParameters.keys(): trainParam[key]=trainParameters[key]
        if trainParam['bandgap'] not in ['all','upper','lower']: raise ValueError('bandgap parameter must be all, lower or upper')
        trainKey = targetCol+"_"+trainParam['bandgap']
        self.logTrain[trainKey]={
            'target_col':targetCol,
            'prediction_type': 'regression',
            'train_parameters':trainParam,
        }
        dfAll =  self.dataset.copy(deep=True)
        for col in self.dataset.columns:
            if col not in ["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap']: dfAll[col]=np.log10(dfAll[col])

        #   Log parameters
        if self.parameters['logML']: self.logger.open()
        Logger.printTitle('TRAINING-REG_'+trainKey,titleLen=60, newLine=False)
        Logger.printTitle('PARAMETERS',titleLen=40)
        Logger.printDic(trainParam)
        if self.parameters['logML']: self.logger.close()

        #   Normalize dataset
        if trainParam['normalize']:
            scaler_dict = {}
            for col in self.dataset.columns:
                if col in ['Name','bandgap']: continue
                scaler_dict[col]=MinMaxScaler()
                # if col not in ["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap']: dfAll[col]=np.log10(dfAll[col])
                dfAll[col]=scaler_dict[col].fit_transform(dfAll[col].values.reshape(-1,1))
        else:
            scaler_dict=None
        self.dfAllnorm=dfAll.copy(deep=True)
        self.logTrain[trainKey]['scaler']=scaler_dict
        #   Prepare Dataset for training
        if  trainParam['bandgap'] == 'upper': dfAll = dfAll.loc[dfAll['bandgap']==1]
        if  trainParam['bandgap'] == 'lower': dfAll = dfAll.loc[dfAll['bandgap']==0]
        dfTrain, dfVal = train_test_split(dfAll, test_size=trainParam['validation_fraction'],random_state=trainParam['random_seed'])
        xTrain = dfTrain.drop(["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap'],axis =1)
        yTrain = dfTrain[targetCol]
        xVal = dfVal.drop(["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap'],axis =1)
        yVal = dfVal[targetCol]

        #   Train model
        if self.parameters['logML']: self.logger.open()
        Logger.printTitle('VERBOSE',titleLen=40)
        model = trainParam['base_model']
        training_start_time = time.time()
        model.fit(xTrain,yTrain)
        training_end_time =time.time()
        self.logTrain[trainKey]['model'] = model
        if self.parameters['logML']: self.logger.close()

        #   Score
        actTrain = yTrain
        predTrain = model.predict(xTrain)
        actVal = yVal
        predVal = model.predict(xVal)

        if trainParam['normalize']:
            actTrain = scaler_dict[targetCol].inverse_transform(actTrain.values.reshape(-1, 1))
            predTrain = scaler_dict[targetCol].inverse_transform(predTrain.reshape(-1, 1))
            actVal = scaler_dict[targetCol].inverse_transform(actVal.values.reshape(-1, 1))
            predVal = scaler_dict[targetCol].inverse_transform(predVal.reshape(-1, 1))

        self.logTrain[trainKey]['results'] = {
            "training_time": "{:.2f} s".format(training_end_time-training_start_time),
            "training_r2": "{:.3f}".format(r2_score(actTrain,predTrain)),
            "validation_r2": "{:.3f}".format(r2_score(actVal,predVal)),
            "training_rmse":"{:.2e}".format(np.sqrt(mean_squared_error(actTrain,predTrain))),
            "validation_rmse":"{:.2e}".format(np.sqrt(mean_squared_error(actVal,predVal))),
        }

        #   Log results
        if self.parameters['logML']: self.logger.open()
        Logger.printTitle('RESULTS',titleLen=40)
        Logger.printDic(self.logTrain[trainKey]['results'])
        if self.parameters['logML']: self.logger.close()

        #   Save validation data
        dfVal_output = dfVal.copy(deep=True)
        if trainParam['normalize']:
            for col in dfVal.columns:
                if col in ['Name','bandgap']: continue
                dfVal_output[col]=scaler_dict[col].inverse_transform(dfVal[col].values.reshape(-1,1))
        dfVal_output['actual'] = actVal
        dfVal_output['predicted'] = predVal
        self.logTrain[trainKey]['validation_data']= dfVal_output
    def trainClassifier(self, targetCol, trainParameters=None):
        '''
        ---Doc---
            Description:
                training pipeline for classifiers. Data and results are stored on the ML object
            Inputs:
                targetCol       string      column name of the dataframe to classify
                trainParameters dicitionary training parameters to overwrite
            Outputs:
                None
        '''
        #   define and update training parameters:
        trainParam={
            'validation_fraction': 0.01,    # validation dataset percentage
            'normalize': True,     # Wether or not to normalize the input data (Must be True for NN)
            'base_model': MLPClassifier((100,100),alpha=0.001, activation = 'relu', learning_rate='adaptive', verbose=0),
            'random_seed': np.random.randint(1000),
            'bandgap': 'all', #or 'upper' == Et>0 or 'lower' ==Et<0
            }
        if trainParameters!=None:
            for key in trainParameters.keys(): trainParam[key]=trainParameters[key]
        if trainParam['bandgap'] not in ['all','upper','lower']: raise ValueError('bandgap parameter must be all, lower or upper')
        trainKey = targetCol+"_"+trainParam['bandgap']
        self.logTrain[trainKey]={
            'target_col':targetCol,
            'prediction_type': 'classification',
            'train_parameters':trainParam,
        }

        dfAll =  self.dataset.copy(deep=True)
        for col in self.dataset.columns:
            if col not in ["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap']: dfAll[col]=np.log10(dfAll[col])

        #   Normalize dataset
        if trainParam['normalize']:
            scaler_dict = {}
            for col in self.dataset.columns:
                if col in ['Name','bandgap']: continue
                scaler_dict[col]=MinMaxScaler()
                dfAll[col]=scaler_dict[col].fit_transform(dfAll[col].values.reshape(-1,1))
        else:
            scaler_dict=None

        self.logTrain[trainKey]['scaler']=scaler_dict

        #   Log parameters
        if self.parameters['logML']: self.logger.open()
        Logger.printTitle('TRAINING-CLASS_'+trainKey,titleLen=60, newLine=False)
        Logger.printTitle('PARAMETERS',titleLen=40)
        Logger.printDic(trainParam)
        if self.parameters['logML']: self.logger.close()

        #   Prepare Dataset for training
        dfTrain, dfVal = train_test_split(dfAll, test_size=trainParam['validation_fraction'],random_state=trainParam['random_seed'])
        xTrain = dfTrain.drop(["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap'],axis =1)
        yTrain = dfTrain[targetCol]
        xVal = dfVal.drop(["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap'],axis =1)
        yVal = dfVal[targetCol]

        #   Train model
        if self.parameters['logML']: self.logger.open()
        Logger.printTitle('VERBOSE',titleLen=40)
        model = trainParam['base_model']
        training_start_time = time.time()
        model.fit(xTrain,yTrain)
        training_end_time =time.time()
        self.logTrain[trainKey]['model'] = model
        if self.parameters['logML']: self.logger.close()

        #   Score
        actTrain = yTrain
        predTrain = model.predict(xTrain)
        actVal = yVal
        predVal = model.predict(xVal)
        probaVal = model.predict_proba(xVal)

        self.logTrain[trainKey]['results'] = {
            "training_time": "{:.2f} s".format(training_end_time-training_start_time),
            "training_logloss": "{:.2e}".format(log_loss(actTrain,predTrain)),
            "validation_logloss": "{:.2e}".format(log_loss(actVal,predVal)),
            "training_accuracy":"{:.3f}".format(accuracy_score(actTrain,predTrain)),
            "validation_accuracy":"{:.3f}".format(accuracy_score(actVal,predVal)),
            "training_f1score":"{:.3f}".format(f1_score(actTrain,predTrain)),
            "validation_f1score":"{:.3f}".format(f1_score(actVal,predVal)),
            "training_precision":"{:.3f}".format(precision_score(actTrain,predTrain)),
            "validation_precision":"{:.3f}".format(precision_score(actVal,predVal)),
            "training_recall":"{:.3f}".format(recall_score(actTrain,predTrain)),
            "validation_recall":"{:.3f}".format(recall_score(actVal,predVal)),
        }
        CM_labels = sorted(self.dataset[targetCol].unique())
        ind = ['Pred_'+str(c) for c in CM_labels]
        col = ['Act_'+str(c) for c in CM_labels]
        self.logTrain[trainKey]['classification_report'] = classification_report(actVal,predVal, digits=3)
        self.logTrain[trainKey]['confusion_matrix'] = pd.DataFrame(confusion_matrix(actVal,predVal),columns=col,index=ind).transpose()
        #   Log results
        if self.parameters['logML']: self.logger.open()
        Logger.printTitle('RESULTS',titleLen=40)
        Logger.printDic(self.logTrain[trainKey]['results'])
        Logger.printTitle('CONFUSION MATRIX',titleLen=40)
        self.printConfusionMatrix(trainKey)
        Logger.printTitle('CLASSIFICATION REPORT',titleLen=40)
        self.printClassificationReport(trainKey)
        if self.parameters['logML']: self.logger.close()

        #   Save validation data
        dfVal_output = dfVal.copy(deep=True)
        if trainParam['normalize']:
            for col in dfVal.columns:
                if col in ['Name','bandgap']: continue
                dfVal_output[col]=scaler_dict[col].inverse_transform(dfVal[col].values.reshape(-1,1))
        dfVal_output['actual'] = actVal
        dfVal_output['predicted'] = predVal
        dfVal_output['predicted_proba'] = [np.max(proba) for proba in probaVal]
        self.logTrain[trainKey]['validation_data']= dfVal_output

    #****   Plotting methods      ****#
    def plotRegressor(self,trainKey, plotParameters=None):
        '''
        ---Doc---
            Description:
                plot true vs predicted on the validation dataset post training
            Inputs:
                trainKey        string      training results to plot
                plotParameters  dicitionary plotting parameters to overwrite
            Outputs:
                None
        '''
        if self.logTrain[trainKey]['prediction_type'] != 'regression': raise ValueError('Wrong prediction type')
        plotParam={
            'figsize':(8,8),
            'xlabel':'True value',
            'ylabel':'Predicted value',
            'show_yx':True,
            'log_plot':False,
            'scatter_alpha':0.8,
            'scatter_s':15,
            'scatter_c': 'C8',
            'save':False,
        }
        if self.parameters['save']: plotParam['save']=True
        if plotParameters!=None:
            for key in plotParameters.keys(): plotParam[key]=plotParameters[key]

        targetCol, bandgapParam = trainKey.rsplit('_',1)
        df_Val = self.logTrain[trainKey]['validation_data']
        results = self.logTrain[trainKey]['results']
        plt.figure(figsize=plotParam['figsize'])
        ax = plt.gca()
        ax.set_xlabel(plotParam['xlabel'])
        ax.set_ylabel(plotParam['ylabel'])
        ax.set_aspect('equal')

        ax.annotate("$\it{R^2}=$%s"%(results['validation_r2']),xy=(0.05,0.95),xycoords='axes fraction', fontsize=14)
        ax.set_title(targetCol+ " on "+bandgapParam+" bandgap", fontsize=14)
        ax.scatter(df_Val.actual,df_Val.predicted,marker=".",alpha=plotParam['scatter_alpha'],s=plotParam['scatter_s'],c=plotParam['scatter_c'])
        if plotParam['show_yx']: ax.plot([np.min([df_Val.actual.min(),df_Val.predicted.min()]),np.max([df_Val.actual.max(),df_Val.predicted.max()])],[np.min([df_Val.actual.min(),df_Val.predicted.min()]),np.max([df_Val.actual.max(),df_Val.predicted.max()])],'k--')
        if plotParam['save']:   plt.savefig(self.pathDic['figures']+datetime.datetime.now().strftime("%Y-%m-%d-%H-%M")+"Reg_"+trainKey+".png",transparent=True,bbox_inches='tight')
        plt.show()
    def printConfusionMatrix(self,trainKey, printParameters=None):
        '''
        ---Doc---
            Description:
                function to print out confusion matrix if it exists
            Inputs:
                trainKey        string      training to print
                printParameters dicitionary print parameters to overwrite
            Outputs:
                None
        '''
        if self.logTrain[trainKey]['prediction_type'] != 'classification': raise ValueError('Wrong prediction type')
        print(self.logTrain[trainKey]['confusion_matrix'])
        print('\n')
    def printClassificationReport(self,trainKey,printParameters=None):
        '''
        ---Doc---
            Description:
                function to print out classification report if it exists
            Inputs:
                trainKey        string      training to print
                printParameters dicitionary print parameters to overwrite
            Outputs:
                None
        '''
        if self.logTrain[trainKey]['prediction_type'] != 'classification': raise ValueError('Wrong prediction type')
        print(self.logTrain[trainKey]['classification_report'])
        print('\n')

    #****   Exporting methods      ****#



"Get attributes and print all attributes"
import pickle

def PrintSelf(obj):
    """Print out objects attribute"""
    for attr in dir(obj):
        if not attr.startswith("__"):
            value = getattr(obj, attr)
            if not callable(value):
                print(str(attr)+" : "+str(value))

def SaveObj(obj, folder, name):
    """Save python object using pickle"""
    if '.pkl' in name:
        with open(folder + name, 'rb') as f:
            pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
    else:
        with open(folder + name + '.pkl', 'wb') as f:
            pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

def LoadObj(folder, name):
    """Load python object using pickle"""
    if '.pkl' in name:
        with open(folder + name, 'rb') as f:
            return pickle.load(f)
    else:
        with open(folder + name + '.pkl', 'rb') as f:
            return pickle.load(f)



"Experimental main functions"
from ..si import Cell,Defect,LTS
from ..main import ML
from ..utils.matplotlibstyle import *
from ..utils import SaveObj, LoadObj
from ..utils import Logger
import numpy as np
import os
import warnings
import pandas as pd
import datetime
import matplotlib.pyplot as plt
import warnings
# warnings.filterwarnings("ignore")

class Experiment():
    #****   constant declaration    ****#
    DefaultParameters = {
        'name':"",  # string added to every saved file for reference
        'save': False,  # True to save a copy of the printed log, the outputed model and data
        'logML': False, # Log copy of print to console into text file
        'n_defects': 100, # Size of simulated defect data set for machine learning
        'dn_range' : np.logspace(13,17,10),# Number of points to interpolate the curves on
        'Et_min':-0.55,  #   Minimum energy level
        'Et_max':0.55,  #   maximum energy level
        'S_min':1E-18,   #   minimum capture cross section
        'S_max':1E-12,  #   maximum capture cross section
        'Nt':1E12,  #   maximum energy level
        'noise_model':"",  #   Type of noise in SRH generation
        'noise_parameter':0, #Parameter used to vary noise level from noise model
        'check_auger':True,     #   Check if lifetime generation should be below Auger limit

    }

    #****   general methods     ****#
    def __init__(self,SaveDir,Parameters=None):
        '''
        ---Doc---
            Description:
                Initialize exp object with passed or default parameters
            Inputs:
                SaveDir     string      Folder path to save the data if parameters['save'] is true
                Parameters  dicitionary Force or overwrite default parameters
            Outputs:
                None
        '''
        #   Check applicability of method
        if not os.path.exists(SaveDir): raise ValueError('%s does not exists'%(SaveDir))

        #   Create directory for computation on this dataset
        self.pathDic = {
            'savedir':      SaveDir,
            'figures':      SaveDir+"\\figures\\",
            'objects':      SaveDir+"\\objects\\",
            'traces':       SaveDir+"\\traces\\",
            'outputs':      SaveDir+"\\outputs\\",
        }
        for key, value in self.pathDic.items():
                if key in ['figures','objects','traces','outputs']:
                    if not os.path.exists(value):   os.makedirs(value)

        #   define hyper parameters for experiment
        self.parameters = Experiment.DefaultParameters
        self.logbook = {'created': datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S")}
        self.logDataset = None
        self.logML = None
        if Parameters is not None: self.updateParameters(Parameters)
    def loadExp(path,filename=None):
        '''
        ---Doc---
            Description:
                Load exp object. If no filename is passed, will loaded latest from saveDir
            Inputs:
                filename    string      file name to load in SaveDir
            Outputs:
                exp     Object
        '''
        if filename != None:
            exp = LoadObj(path,filename)
            exp.updateLogbook('Experiment_loaded_'+filename)
            return(exp)
        else:   # if no filename provided, take the latest one, if none exists, raise error.
            current_timestamp = datetime.datetime(1990, 10, 24, 16, 00, 00)
            for file in os.scandir(path):
                if not file.is_file(): continue
                if 'experimentObj' in file.name:
                    timestamp = datetime.datetime.strptime(file.name.split("_")[-1].split(".")[0],"%Y-%m-%d-%H-%M-%S")
                    if timestamp > current_timestamp:
                        filename = file.name
                        current_timestamp=timestamp
            if filename != None:
                exp = LoadObj(path,filename)
                exp.updateLogbook('Experiment_loaded_'+filename)
                return(exp)
            else:
                raise ValueError("No experimental file exists in %s"%(path))
    def saveExp(self, name=None):
        '''
        ---Doc---
            Description:
                Save exp object with pickle
            Inputs:
                name    string  Overwrite filename
            Outputs:
                None
        '''
        if name == None: name = self.parameters['name']
        # if self.logML !=None:
        #     for id,ML in self.logML:
        #         savePath=
        self.updateLogbook('Experiment_saved_'+name)
        SaveObj(self,self.pathDic['objects'],'experimentObj_'+name+"_"+datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S"))

    #****   machine learning methods     ****#
    def newML(self, datasetID=None, mlParameters=None):
        '''
        ---Doc---
            Description:
                Define new ML object based on parameters from the experiment
            Inputs:
                datasetID       string      Overwrite which dataset to use from logDataset
                mlParameters    dicitionary Overwrite ml parameters for object creation
            Outputs:
                mlID     string mlID of newly created object
        '''
        if self.logDataset == None : raise ValueError("Experiment doesn't have simulated data.")
        if datasetID==None: datasetID = str(len(self.logDataset)-1) # use latest if not defined
        if not isinstance(datasetID,str): datasetID = str(datasetID)
        mlParam = {'name':None,'save':None,'logML':None}
        if mlParameters!=None:
            for key in mlParameters.keys(): mlParam[key]=mlParameters[key]
        if mlParam['save']==None: mlParam['save']=self.parameters['save']
        if mlParam['logML']==None: mlParam['logML']=self.parameters['logML']
        if self.logML==None:
            mlID = "0"
        else:
            mlID = str(len(self.logML))
        if mlParam['name']==None:mlParam['name']= self.parameters['name']+"-#"+mlID
        mlParam['mlID']=mlID
        ml = ML(Dataset=self.logDataset[datasetID], SaveDir=self.pathDic['savedir'],Parameters=mlParam)

        #   Log change
        mlID=self.updateLogMLmodel(ml, logID=mlID)
        self.updateLogbook('ML_model_created_ID'+mlID+"_"+ml.parameters['name'])
        return(self.logML[mlID])
    def loadML(self, filename=None):
        '''
        ---Doc---
            Description:
                Load ml object. If no filename is passed, will loaded latest from saveDir
            Inputs:
                filename    string      file name to load in SaveDir
            Outputs:
                ml     Object
        '''
        if filename != None:
            ml = LoadObj(self.pathDic['objects'],filename)
        else:   # if no filename provided, take the latest one, if none exists, raise error.
            current_timestamp = datetime.datetime(1990, 10, 24, 16, 00, 00)
            for file in os.scandir(self.pathDic['objects']):
                if not file.is_file(): continue
                if 'mlObj' in file.name:
                    timestamp = datetime.datetime.strptime(file.name.split("_")[-1].split(".")[0],"%Y-%m-%d-%H-%M-%S")
                    if timestamp > current_timestamp:
                        filename = file.name
                        current_timestamp=timestamp
            if filename != None:
                ml = LoadObj(self.pathDic['objects'],filename)
            else:
                raise ValueError("No ltsDF file exists in %s"%(self.pathDic['objects']))

        if ml.parameters['mlID'] != None:
            self.updateLogMLmodel(ml,logID=ml.parameters['mlID'])
        else:
            mlID = self.updateLogMLmodel(ml)
            ml.parameters['mlID'] =mlID
        self.updateLogbook('ML_loaded_ID'+ml.parameters['mlID']+"_from_"+filename)
        if ml.logger !=None: ml.logger = Logger(ml.logger)
        return(ml)
    def predictML(self, mlIDs = None, header = None):
        '''
        ---Doc---
            Description:
                Apply ml object regressor and classifier to experimental data and save results locally
            Inputs:
                mlIDs   array      list of mlID string to use for prediction. All by defaultself.
                header  array       column names of predicted value to scale
            Outputs:
                None
        '''        #   Check for applicabiliy
        if mlIDs == None: mlIDs = [str(i) for i in range(len(self.logML))]
        if header == None: header = ['Et_eV','Sn_cm2','Sp_cm2','k','logSn','logSp','logk']
        self.predictCsv = {}
        for mlID in mlIDs:
            ml = self.logML[mlID]
            self.predictCsv[mlID]={}
            #   Create dataset feature vector
            for trainKey, mlDic in ml.logTrain.items():
                vector = [t for key in self.expKeys for t in self.expDic[key]['tau_interp']]
                if trainKey=='scaler': continue
                targetCol, bandgapParam = trainKey.rsplit('_',1)
                if mlDic['train_parameters']['normalize']:
                    #   scale the feature vector
                    if len(vector) != len(mlDic['scaler'])-len(header): raise ValueError('Feature vector is not the same size as the trained ML')
                    i=0
                    for scaler_key,scaler_value in mlDic['scaler'].items():
                        if scaler_key in header: continue
                        vector[i]=np.log10(vector[i])
                        vector[i]=scaler_value.transform(vector[i].reshape(-1,1))[0][0]
                        i+=1
                #   Call ML model and predict on sample data
                if mlDic['prediction_type'] == 'regression':
                    if mlDic['train_parameters']['normalize']:  self.predictCsv[mlID][trainKey] = mlDic['scaler'][targetCol].inverse_transform([mlDic['model'].predict([vector])])[0][0]
                if mlDic['prediction_type'] == 'classification':
                    self.predictCsv[mlID][trainKey] = (mlDic['model'].predict([vector])[0],mlDic['model'].predict_proba([vector])[0])

            #   Log in the trace
            if ml.parameters['logML']: ml.logger.open()
            Logger.printTitle(' ML PREDICTION',titleLen=60, newLine=False)
            Logger.printTitle('mlID '+mlID,titleLen=40)
            Logger.printDic(self.predictCsv[mlID])
            if ml.parameters['logML']: ml.logger.close()

        self.updateLogbook('prediction_made')

    #****   simulation methods     ****#
    def generateDB(self):
        '''
        ---Doc---
            Description:
                Generate defect database, from object parameters. Each database generated will have a separate id.
            Inputs:
                None
            Outputs:
                None
        '''
        # Generate Random defect database
        defectDB=Defect.randomDB(
            N=self.parameters['n_defects'],
            Et_min = self.parameters['Et_min'],
            Et_max = self.parameters['Et_max'],
            S_min = self.parameters['S_min'],
            S_max = self.parameters['S_max'],
            Nt = self.parameters['Nt']
        )

        # Calculate cell level characteristic for temperature,doping
        cref = Cell(T=300,Ndop=1E15,type=self.parameters['type'])
        cellDB = [cref.changeT(T).changeNdop(Ndop) for (T,Ndop) in zip(self.parameters['temperature'],self.parameters['doping'])]

        # Calculate lifetime on database and check for Auger limit
        columns_name = ["Name","Et_eV","Sn_cm2","Sp_cm2",'k','logSn','logSp','logk','bandgap']
        ltsDB=[]
        firstPass = True
        noiseparam = 0
        for d in defectDB:
            bandgap = 1 if d.Et>0 else 0
            col = [d.name,d.Et,d.Sn,d.Sp,d.k, np.log10(d.Sn),np.log10(d.Sp),np.log10(d.k),bandgap]
            skipDefect = False
            for c in cellDB:
                if firstPass:
                    for dn in self.parameters['dn_range']: columns_name.append("%sK_%scm-3_ %scm-3" % (c.T,c.Ndop,dn))
                    skipDefect = True
                    continue
                if skipDefect: continue
                s = LTS(c,d,self.parameters['dn_range'],noise=self.parameters['noise_model'], noiseparam=self.parameters['noise_parameter'])
                if self.parameters['check_auger']:  #if break auger limit, discard defect
                    breakAuger,_ = s.checkAuger()
                    if breakAuger: skipDefect=True
                if skipDefect: continue
                for t,dn in zip(s.tauSRH_noise,s.dnrange):
                    col.append(t)
            firstPass = False
            if not skipDefect: ltsDB.append(col)
            if skipDefect:  # if we skipped a defect, add a new random one to have n_defects in the database at the end
                newDefect = Defect.randomDB(
                        N=1,
                        Et_min = self.parameters['Et_min'],
                        Et_max = self.parameters['Et_max'],
                        S_min = self.parameters['S_min'],
                        S_max = self.parameters['S_max'],
                        Nt = self.parameters['Nt']
                        )[0]
                newDefect.name = d.name
                defectDB.append(newDefect)
        ltsDF = pd.DataFrame(ltsDB)
        ltsDF.columns = columns_name
        ltsID = self.updateLogDataset(ltsDF)
        if self.parameters['save']:
            SaveObj(ltsDF,self.pathDic['objects'],'ltsDF_ID'+ltsID+"_"+datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S"))
            self.updateLogbook('lifetime_database_saved_ID'+ltsID)

        #   Log change
        self.updateLogbook('lifetime_database_generated_ID'+ltsID)
    def interpolateSRH(self):
        '''
        ---Doc---
            Description:
                After loading experimental data, linearize and interpolates SRH curves
            Inputs:
                None
            Outputs:
                None
        '''
        #   For each SRH curve, linearize and interpolate on dn_range
        for key, curve in self.expDic.items():
            c = curve['cell']
            curve['X_linear_csv']=[(1)/(c.n0+c.p0+dn) for dn in curve['dn_csv']]
            curve['X_linear_interp']=[(1)/(c.n0+c.p0+dn) for dn in self.parameters['dn_range']]
            curve['Fit_slope'],curve['Fit_intercep'] = np.polyfit(curve['X_linear_csv'],curve['tau_csv'],deg=1)
            curve['tau_interp']=[curve['Fit_slope']*X+curve['Fit_intercep'] for X in curve['X_linear_interp']]

        #   Log change
        self.updateLogbook('interpolated')
    def loadCSV(self,FilePath, Temperature, Doping, Type):
        '''
        ---Doc---
            Description:
                Load csv file of experimental data and prep data. Specify experimental condition with Temperature, Doping and Type
            Inputs:
                FilePath    string  .csv file where data is stored. use sample.csv for formatting
                Temperature array   array of temperatures used in measurements
                Doping      array   array of doping used in measurements. Need to be the same length as Temperature
                Type        string  n or p type
            Outputs:
                None
        '''
        #   Check applicability of method
        if not os.path.exists(FilePath): raise ValueError('%s does not exists'%(FilePath))
        if '.csv' not in FilePath: raise ValueError('%s is not a csv file'%(FilePath))
        if len(Temperature)!=len(Doping): raise ValueError('Doping and Temperature array needs to be the same length')
        if Type not in ['n','p']: raise ValueError('%s needs to be either n or p'%(Type))

        #   Load csv file and check the correct format
        self.pathDic['csvfile']=FilePath
        self.csvDf = pd.read_csv(FilePath)
        if len(self.csvDf.columns)!=2*len(Temperature): raise ValueError('csv file does not match to Temperature and Doping array given')

        #   Create experimental data storage for further calculations
        self.expKeys = ["Exp#"+str(i) for i in range(len(Temperature))]
        new_columns = [[k1+"_Dn",k2+"_Tau"] for k1,k2 in zip(self.expKeys,self.expKeys)]
        self.csvDf.columns = [l1 for subL in new_columns for l1 in subL]
        self.expDic={}
        for key,T,Ndop in zip(self.expKeys,Temperature,Doping):
            self.expDic[key]={
                'T':T,
                'Ndop':Ndop,
                'dn_csv':self.csvDf[key+"_Dn"].dropna().values,
                'tau_csv':self.csvDf[key+"_Tau"].dropna().values,
                'cell': Cell(T=T,Ndop=Ndop,type=Type),
            }

        #   Save parameter in Parameter dicitionary
        changedParameter = {'type':Type,'temperature':Temperature,'doping':Doping}
        self.updateParameters(changedParameter)
        #   Log change
        self.updateLogbook('csv_loaded')
    def loadLTS(self, filename=None):
        '''
        ---Doc---
            Description:
                load previously generated lifetime data
            Inputs:
                filename    string      file name to load in SaveDir
            Outputs:
                None
        '''
        if filename != None:
            ltsDF = LoadObj(self.pathDic['objects'],filename)
            ltsID = self.updateLogDataset(ltsDF)
            self.updateLogbook('ltsDB_loaded_ID'+ltsID+"_from_"+filename)
        else:   # if no filename provided, take the latest one, if none exists, raise error.
            current_timestamp = datetime.datetime(1990, 10, 24, 16, 00, 00)
            for file in os.scandir(self.pathDic['objects']):
                if not file.is_file(): continue
                if 'ltsDF' in file.name:
                    timestamp = datetime.datetime.strptime(file.name.split("_")[-1].split(".")[0],"%Y-%m-%d-%H-%M-%S")
                    if timestamp > current_timestamp:
                        filename = file.name
                        current_timestamp=timestamp
            if filename != None:
                ltsDF = LoadObj(self.pathDic['objects'],filename)
                ltsID = self.updateLogDataset(ltsDF)
                self.updateLogbook('ltsDB_loaded_ID'+ltsID+"_from_"+filename)
            else:
                raise ValueError("No ltsDF file exists in %s"%(self.pathDic['objects']))

    #****   plotting methods     ****#
    def plotSRH(self,toPlot=None, plotParameters=None):
        '''
        ---Doc---
            Description:
                plot SRH lifetime curve and/or interpolated SRH curves
            Inputs:
                toPlot          array      list of options to plot
                plotParameters  dicitionary plotting parameters to overwrite
            Outputs:
                None
        '''
        #   Check applicability of method
        if toPlot==None:
            toPlot=[]
            if 'csv_loaded' in self.logbook.keys(): toPlot.append('fromCSV')
            if 'interpolated' in self.logbook.keys(): toPlot.append('fromInterpolated')

        #   define and update plot parameters:
        plotParam={
            'figsize':(8,8),
            'colorscale':plt.cm.RdYlBu_r(np.linspace(0.1,0.9,len(self.expKeys))),
            'save':self.parameters['save'],
            'xlabel':'Excess minority carrier [cm$^{-3}$]',
            'ylabel':'Lifetime [s]',
            'legend':True,
            'label_fromCSV':['Experimental - (%.0F, %.1E)'%(self.expDic[key]['T'],self.expDic[key]['Ndop']) for key in self.expKeys],
            'label_fromInterpolated':['Interpolation - (%.0F, %.1E)'%(self.expDic[key]['T'],self.expDic[key]['Ndop']) for key in self.expKeys],
            'xrange':[np.min(self.parameters['dn_range']),np.max(self.parameters['dn_range'])],
            'yrange':None,
        }
        if plotParameters!=None:
            for key in plotParameters.keys(): plotParam[key]=plotParameters[key]

        #   plot figure
        plt.figure(figsize=plotParam['figsize'])
        ax = plt.gca()
        ax.set_xlabel(plotParam['xlabel'])
        ax.set_ylabel(plotParam['ylabel'])
        k=0
        ymin,ymax=np.infty,0
        for key in self.expKeys:
            if 'fromCSV' in toPlot:
                ax.scatter(self.expDic[key]['dn_csv'],self.expDic[key]['tau_csv'],c=plotParam['colorscale'][k],label=plotParam['label_fromCSV'][k])
                if np.min(self.expDic[key]['tau_csv'])<ymin: ymin = np.min(self.expDic[key]['tau_csv'])
                if np.max(self.expDic[key]['tau_csv'])>ymax: ymax = np.max(self.expDic[key]['tau_csv'])
            if 'fromInterpolated' in toPlot:
                ax.plot( self.parameters['dn_range'],self.expDic[key]['tau_interp'],c=plotParam['colorscale'][k],label=plotParam['label_fromInterpolated'][k])
                if np.min(self.expDic[key]['tau_interp'])<ymin: ymin = np.min(self.expDic[key]['tau_interp'])
                if np.max(self.expDic[key]['tau_interp'])>ymax: ymax = np.max(self.expDic[key]['tau_interp'])
            k+=1
        if plotParam['yrange']==None:  plotParam['yrange']=[0.9*ymin,1.1*ymax]
        if plotParam['xrange']!=None: ax.set_xlim(left=plotParam['xrange'][0],right=plotParam['xrange'][1])
        if plotParam['yrange']!=None: ax.set_ylim(bottom=plotParam['yrange'][0],top=plotParam['yrange'][1])
        if plotParam['legend']: ax.legend(ncol=2,bbox_to_anchor=(1,0.5), loc='center left')
        ax.loglog()
        if plotParam['save']: plt.savefig(self.pathDic['figures']+"plotSRH"+datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")+self.parameters['name']+".png",transparent=True,bbox_inches='tight')
        plt.show()

    #****   exporting methods     ****#
    def exportDataset(self,filename=None):
        '''
        ---Doc---
            Description:
                Convert datasets used into csv files
            Inputs:
                name    string  Overwrite filename
            Outputs:
                None
        '''
        if filename == None: filename = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")+"_"+self.parameters['name']
        for key, dataset in self.logDataset.items():
            filename_temp=filename+"_datasetID_"+key+".csv"
            dataset.to_csv(self.pathDic['outputs']+filename_temp,encoding='utf-8', index=False)
        self.updateLogbook('dataset_exported')
    def exportValidationset(self,filename=None):
        '''
        ---Doc---
            Description:
                Convert validation datasets used into csv files
            Inputs:
                name    string  Overwrite filename
            Outputs:
                None
        '''
        if filename == None: filename = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")+"_"+self.parameters['name']
        for mlID, ml in self.logML.items():
            for trainKey, trainLog in ml.logTrain.items():
                filename_temp = filename+"_mlID_"+mlID+"_ValidationSet_"+trainKey+'.csv'
                trainLog['validation_data'].to_csv(self.pathDic['outputs']+filename_temp,encoding='utf-8', index=False)
        self.updateLogbook('validation_set_exported')

    #****   updating methods      ****#
    def updateParameters(self,Parameters):
        '''
        ---Doc---
            Description:
                update objects parameter dicitionary
            Inputs:
                Parameters  dicitionary     parameters to overwrite
            Outputs:
                None
        '''
        for key,value in Parameters.items():
            self.parameters[key]=value
    def updateLogbook(self,logItem):
        '''
        ---Doc---
            Description:
                update logbook with item and timestamp
            Inputs:
                logItem  string item name to add to logbook
            Outputs:
                None
        '''
        self.logbook[logItem]=datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S")
    def updateLogDataset(self,logItem):
        '''
        ---Doc---
            Description:
                update dataset log with dataset
            Inputs:
                logItem  string item name to add to log
            Outputs:
                id  string  highest id found in log
        '''
        if self.logDataset==None:
            self.logDataset={"0":logItem}
            id = "0"
        else:
            found = False
            for key,value in self.logDataset.items():
                if logItem.equals(value):  id, found = key, True
            if not found:
                id = str(len(self.logDataset))
                self.logDataset[id]=logItem
        return(id)
    def updateLogMLmodel(self,logItem,logID=None):
        '''
        ---Doc---
            Description:
                update dataset log with ML model
            Inputs:
                logItem  string item name to add to log
                logID    string overwrite logID to use
            Outputs:
                id  string  highest id found in log
        '''
        if self.logML==None:
            self.logML={"0":logItem}
            id = "0"
        else:
            if logID==None:
                id = str(len(self.logML))
            else:
                id = logID
            self.logML[id]=logItem
        return(id)
